from pyiron_lammps.decorator import calculation
from atomistics.workflows import (
    ElasticMatrixWorkflow,
    EnergyVolumeCurveWorkflow,
    optimize_positions_and_volume,
)


def _run_simulation(structure, potential_dataframe, input_template, lmp):
    # write structure to LAMMPS
    lmp.interactive_structure_setter(
        structure=structure,
        units="metal",
        dimension=3,
        boundary=" ".join(["p" if coord else "f" for coord in structure.pbc]),
        atom_style="atomic",
        el_eam_lst=potential_dataframe.Species,
        calc_md=False,
    )

    # execute calculation
    for c in potential_dataframe.Config:
        lmp.interactive_lib_command(c)

    for l in input_template.split("\n"):
        lmp.interactive_lib_command(l)

    return lmp


def _optimize_structure_optional(
    lmp, structure, potential_dataframe, minimization_activated=True
):
    if minimization_activated:
        return optimize_structure(
            lmp=lmp, structure=structure, potential_dataframe=potential_dataframe
        )
    else:
        return structure


@calculation
def optimize_structure(lmp, structure, potential_dataframe):
    from atomistics.calculators import evaluate_with_lammps_library

    task_dict = optimize_positions_and_volume(structure=structure)
    structure_copy = evaluate_with_lammps_library(
        task_dict=task_dict,
        potential_dataframe=potential_dataframe,
        lmp=lmp,
        lmp_optimizer_kwargs={},
    )["structure_with_optimized_positions_and_volume"]

    # clean memory
    lmp.interactive_lib_command("clear")
    return structure_copy


@calculation
def calculate_elastic_constants(
    lmp,
    structure,
    potential_dataframe,
    num_of_point=5,
    eps_range=0.005,
    sqrt_eta=True,
    fit_order=2,
    minimization_activated=False,
):
    lammps_input_template_minimize_pos = """\
variable thermotime equal 100
thermo_style custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol
thermo_modify format float %20.15g
thermo ${thermotime}
min_style cg
minimize 0.0 0.0001 100000 10000000"""

    # Optimize structure
    structure_opt = _optimize_structure_optional(
        lmp=lmp,
        structure=structure,
        potential_dataframe=potential_dataframe,
        minimization_activated=minimization_activated,
    )

    # Generate structures
    calculator = ElasticMatrixWorkflow(
        structure=structure_opt.copy(),
        num_of_point=num_of_point,
        eps_range=eps_range,
        sqrt_eta=sqrt_eta,
        fit_order=fit_order,
    )
    structure_dict = calculator.generate_structures()

    # run calculation
    energy_tot_lst = {}
    for key, struct in structure_dict["calc_energy"].items():
        lmp = _run_simulation(
            lmp=lmp,
            structure=struct,
            potential_dataframe=potential_dataframe,
            input_template=lammps_input_template_minimize_pos,
        )
        energy_tot_lst[key] = lmp.interactive_energy_tot_getter()
        lmp.interactive_lib_command("clear")

    # fit
    calculator.analyse_structures({"energy": energy_tot_lst})
    return calculator._data["C"]


@calculation
def calculate_energy_volume_curve(
    lmp,
    structure,
    potential_dataframe,
    num_points=11,
    fit_type="polynomial",
    fit_order=3,
    vol_range=0.05,
    axes=("x", "y", "z"),
    strains=None,
    minimization_activated=False,
):
    lammps_input_calc_static = """\
variable thermotime equal 100
thermo_style custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol
thermo_modify format float %20.15g
thermo ${thermotime}
run 0"""

    # Optimize structure
    structure_opt = _optimize_structure_optional(
        lmp=lmp,
        structure=structure,
        potential_dataframe=potential_dataframe,
        minimization_activated=minimization_activated,
    )

    # Generate structures
    calculator = EnergyVolumeCurveWorkflow(
        structure=structure_opt.copy(),
        num_points=num_points,
        fit_type=fit_type,
        fit_order=fit_order,
        vol_range=vol_range,
        axes=axes,
        strains=strains,
    )
    structure_dict = calculator.generate_structures()

    # run calculation
    energy_tot_lst = {}
    for key, struct in structure_dict["calc_energy"].items():
        lmp = _run_simulation(
            lmp=lmp,
            structure=struct,
            potential_dataframe=potential_dataframe,
            input_template=lammps_input_calc_static,
        )
        energy_tot_lst[key] = lmp.interactive_energy_tot_getter()
        lmp.interactive_lib_command("clear")

    # fit
    calculator.analyse_structures({"energy": energy_tot_lst})
    return calculator.fit_dict



def calculation(funct):
    def funct_return(lmp=None, *args, **kwargs):
        # Create temporary LAMMPS instance if necessary
        if lmp is None:
            from pylammpsmpi import LammpsASELibrary

            close_lmp_after_calculation = True
            lmp = LammpsASELibrary()
        else:
            close_lmp_after_calculation = False

        # Run function
        result = funct(lmp=lmp, *args, **kwargs)

        # Close temporary LAMMPS instance
        if close_lmp_after_calculation:
            lmp.close()
        return result

    return funct_return



import numpy as np
from ase.atoms import Atoms
from pandas import DataFrame, Series

from pyiron_lammps.calculation import (
    optimize_structure,
    calculate_elastic_constants,
    calculate_energy_volume_curve,
)


def _get_lammps_mpi(enable_mpi=True):
    from pylammpsmpi import LammpsASELibrary

    if enable_mpi:
        # To get the right instance of MPI.COMM_SELF it is necessary to import it inside the function.
        from mpi4py import MPI

        return LammpsASELibrary(
            working_directory=None,
            cores=1,
            comm=MPI.COMM_SELF,
            logger=None,
            log_file=None,
            library=None,
            diable_log_file=True,
        )
    else:
        return LammpsASELibrary(
            working_directory=None,
            cores=1,
            comm=None,
            logger=None,
            log_file=None,
            library=None,
            diable_log_file=True,
        )


def _parallel_execution(function, input_parameter_lst, cores=1):
    if cores == 1:
        return [
            function(input_parameter=input_parameter + [False])
            for input_parameter in input_parameter_lst
        ]
    elif cores > 1:
        from pympipool.mpi import PyMPIExecutor

        with PyMPIExecutor(max_workers=cores) as p:
            return list(
                p.map(
                    function,
                    [
                        input_parameter + [True]
                        for input_parameter in input_parameter_lst
                    ],
                )
            )
    else:
        raise ValueError("The number of cores has to be a positive integer.")


def _optimize_structure_serial(input_parameter):
    structure, potential_dataframe, enable_mpi = input_parameter
    return optimize_structure(
        lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
        structure=structure,
        potential_dataframe=potential_dataframe,
    )


def _calculate_elastic_constants_serial(input_parameter):
    (
        structure,
        potential_dataframe,
        num_of_point,
        eps_range,
        sqrt_eta,
        fit_order,
        minimization_activated,
        enable_mpi,
    ) = input_parameter
    return calculate_elastic_constants(
        lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
        structure=structure,
        potential_dataframe=potential_dataframe,
        num_of_point=num_of_point,
        eps_range=eps_range,
        sqrt_eta=sqrt_eta,
        fit_order=fit_order,
        minimization_activated=minimization_activated,
    )


def _calculate_energy_volume_curve_serial(input_parameter):
    (
        structure,
        potential_dataframe,
        num_points,
        fit_type,
        fit_order,
        vol_range,
        axes,
        strains,
        minimization_activated,
        enable_mpi,
    ) = input_parameter
    return calculate_energy_volume_curve(
        lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
        structure=structure,
        potential_dataframe=potential_dataframe,
        num_points=num_points,
        fit_type=fit_type,
        fit_order=fit_order,
        vol_range=vol_range,
        axes=axes,
        strains=strains,
        minimization_activated=minimization_activated,
    )


def optimize_structure_parallel(structure_list, potential_dataframe_list, cores=1):
    if isinstance(structure_list, (list, np.ndarray)):
        if isinstance(potential_dataframe_list, (list, np.ndarray)):
            if len(structure_list) == len(potential_dataframe_list):
                return _parallel_execution(
                    function=_optimize_structure_serial,
                    input_parameter_lst=[
                        [structure, potential]
                        for structure, potential in zip(
                            structure_list, potential_dataframe_list
                        )
                    ],
                    cores=cores,
                )
            else:
                raise ValueError(
                    "Input lists have len(structure_list) != len(potential_dataframe_list) ."
                )
        elif isinstance(potential_dataframe_list, (DataFrame, Series)):
            return _parallel_execution(
                function=_optimize_structure_serial,
                input_parameter_lst=[
                    [structure, potential_dataframe_list]
                    for structure in structure_list
                ],
                cores=cores,
            )
        else:
            raise TypeError(
                "potential_dataframe_list should either be an pandas.DataFrame object or a list of those. "
            )
    elif isinstance(structure_list, Atoms):
        if isinstance(potential_dataframe_list, (list, np.ndarray)):
            return _parallel_execution(
                function=_optimize_structure_serial,
                input_parameter_lst=[
                    [structure_list, potential]
                    for potential in potential_dataframe_list
                ],
                cores=cores,
            )
        elif isinstance(potential_dataframe_list, (DataFrame, Series)):
            return optimize_structure(
                lmp=_get_lammps_mpi(enable_mpi=False),
                structure=structure_list,
                potential_dataframe=potential_dataframe_list,
            )
        else:
            raise TypeError(
                "potential_dataframe_list should either be an pandas.DataFrame object or a list of those. "
            )
    else:
        raise TypeError(
            "structure_list should either be an ase.atoms.Atoms object or a list of those."
        )


def calculate_elastic_constants_parallel(
    structure_list,
    potential_dataframe_list,
    num_of_point=5,
    eps_range=0.005,
    sqrt_eta=True,
    fit_order=2,
    minimization_activated=False,
    cores=1,
):
    if isinstance(structure_list, (list, np.ndarray)):
        if isinstance(potential_dataframe_list, (list, np.ndarray)):
            if len(structure_list) == len(potential_dataframe_list):
                return _parallel_execution(
                    function=_calculate_elastic_constants_serial,
                    input_parameter_lst=[
                        [
                            structure,
                            potential,
                            num_of_point,
                            eps_range,
                            sqrt_eta,
                            fit_order,
                            minimization_activated,
                        ]
                        for structure, potential in zip(
                            structure_list, potential_dataframe_list
                        )
                    ],
                    cores=cores,
                )
            else:
                raise ValueError(
                    "Input lists have len(structure_list) != len(potential_dataframe_list) ."
                )
        elif isinstance(potential_dataframe_list, (DataFrame, Series)):
            return _parallel_execution(
                function=_calculate_elastic_constants_serial,
                input_parameter_lst=[
                    [
                        structure,
                        potential_dataframe_list,
                        num_of_point,
                        eps_range,
                        sqrt_eta,
                        fit_order,
                        minimization_activated,
                    ]
                    for structure in structure_list
                ],
                cores=cores,
            )
        else:
            raise TypeError(
                "potential_dataframe_list should either be an pandas.DataFrame object or a list of those. "
            )
    elif isinstance(structure_list, Atoms):
        if isinstance(potential_dataframe_list, (list, np.ndarray)):
            return _parallel_execution(
                function=_calculate_elastic_constants_serial,
                input_parameter_lst=[
                    [
                        structure_list,
                        potential,
                        num_of_point,
                        eps_range,
                        sqrt_eta,
                        fit_order,
                        minimization_activated,
                    ]
                    for potential in potential_dataframe_list
                ],
                cores=cores,
            )
        elif isinstance(potential_dataframe_list, (DataFrame, Series)):
            return calculate_elastic_constants(
                lmp=_get_lammps_mpi(enable_mpi=False),
                structure=structure_list,
                potential_dataframe=potential_dataframe_list,
                num_of_point=num_of_point,
                eps_range=eps_range,
                sqrt_eta=sqrt_eta,
                fit_order=fit_order,
                minimization_activated=minimization_activated,
            )
        else:
            raise TypeError(
                "potential_dataframe_list should either be an pandas.DataFrame object or a list of those. "
            )
    else:
        raise TypeError(
            "structure_list should either be an ase.atoms.Atoms object or a list of those."
        )


def calculate_energy_volume_curve_parallel(
    structure_list,
    potential_dataframe_list,
    num_points=11,
    fit_type="polynomial",
    fit_order=3,
    vol_range=0.05,
    axes=["x", "y", "z"],
    strains=None,
    minimization_activated=False,
    cores=1,
):
    if isinstance(structure_list, (list, np.ndarray)):
        if isinstance(potential_dataframe_list, (list, np.ndarray)):
            if len(structure_list) == len(potential_dataframe_list):
                return _parallel_execution(
                    function=_calculate_energy_volume_curve_serial,
                    input_parameter_lst=[
                        [
                            structure,
                            potential,
                            num_points,
                            fit_type,
                            fit_order,
                            vol_range,
                            axes,
                            strains,
                            minimization_activated,
                        ]
                        for structure, potential in zip(
                            structure_list, potential_dataframe_list
                        )
                    ],
                    cores=cores,
                )
            else:
                raise ValueError(
                    "Input lists have len(structure_list) != len(potential_dataframe_list) ."
                )
        elif isinstance(potential_dataframe_list, (DataFrame, Series)):
            return _parallel_execution(
                function=_calculate_energy_volume_curve_serial,
                input_parameter_lst=[
                    [
                        structure,
                        potential_dataframe_list,
                        num_points,
                        fit_type,
                        fit_order,
                        vol_range,
                        axes,
                        strains,
                        minimization_activated,
                    ]
                    for structure in structure_list
                ],
                cores=cores,
            )
        else:
            raise TypeError(
                "potential_dataframe_list should either be an pandas.DataFrame object or a list of those. "
            )
    elif isinstance(structure_list, Atoms):
        if isinstance(potential_dataframe_list, (list, np.ndarray)):
            return _parallel_execution(
                function=_calculate_energy_volume_curve_serial,
                input_parameter_lst=[
                    [
                        structure_list,
                        potential,
                        num_points,
                        fit_type,
                        fit_order,
                        vol_range,
                        axes,
                        strains,
                        minimization_activated,
                    ]
                    for potential in potential_dataframe_list
                ],
                cores=cores,
            )
        elif isinstance(potential_dataframe_list, (DataFrame, Series)):
            return calculate_energy_volume_curve(
                lmp=_get_lammps_mpi(enable_mpi=False),
                structure=structure_list,
                potential_dataframe=potential_dataframe_list,
                num_points=num_points,
                fit_type=fit_type,
                fit_order=fit_order,
                vol_range=vol_range,
                axes=axes,
                strains=strains,
                minimization_activated=minimization_activated,
            )
        else:
            raise TypeError(
                "potential_dataframe_list should either be an pandas.DataFrame object or a list of those. "
            )
    else:
        raise TypeError(
            "structure_list should either be an ase.atoms.Atoms object or a list of those."
        )



from . import _version

from pyiron_lammps.calculation import (
    optimize_structure,
    calculate_elastic_constants,
    calculate_energy_volume_curve,
)
from pyiron_lammps.parallel import (
    optimize_structure_parallel,
    calculate_elastic_constants_parallel,
    calculate_energy_volume_curve_parallel,
)

try:
    from atomistics.calculators.lammps import (
        get_potential_dataframe,
    )
except ImportError:
    pass


def get_lammps_engine(
    working_directory=None,
    cores=1,
    comm=None,
    logger=None,
    log_file=None,
    library=None,
    diable_log_file=True,
):
    from pylammpsmpi import LammpsASELibrary

    return LammpsASELibrary(
        working_directory=working_directory,
        cores=cores,
        comm=comm,
        logger=logger,
        log_file=log_file,
        library=library,
        diable_log_file=diable_log_file,
    )


__version__ = _version.get_versions()["version"]
